/*
 * segmentize.h
 *
 *  Created on: 10 aug 2012
 *      Author: Wolfcastle
 *
 *  Description
 *		This library implements a function that segments
 *      chromosome files into smaller chunks
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

extern int segmentize(int*, char*, char*, int);

